/*
 * Copyright 2016 - 2025 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <ox/std/def.hpp>
#include <ox/std/error.hpp>
#include <ox/std/string.hpp>

#include "asset.hpp"
#include "context.hpp"

namespace keel {

class Wrap {
	public:
		virtual ~Wrap() = default;
		[[nodiscard]]
		virtual ox::CStringView typeName() const noexcept = 0;
		[[nodiscard]]
		virtual int typeVersion() const noexcept = 0;
		[[nodiscard]]
		virtual ox::UAnyPtr moveToCopy() noexcept = 0;
};

template<typename T>
class WrapT: public Wrap {
	public:
		[[nodiscard]]
		virtual constexpr T &obj() noexcept = 0;

		ox::UAnyPtr moveToCopy() noexcept final {
			return new T{std::move(obj())};
		}

};

template<typename T>
class WrapRef final: public WrapT<T> {
	private:
		T &m_obj;

	public:
		constexpr explicit WrapRef(T &obj): m_obj{obj} {}

		ox::CStringView typeName() const noexcept override {
			return ox::ModelTypeName_v<T>;
		}

		int typeVersion() const noexcept override {
			return ox::ModelTypeVersion_v<T>;
		}

		constexpr T &obj() noexcept override {
			return m_obj;
		}

};

template<typename T>
class WrapInline final: public WrapT<T> {
	private:
		T m_obj;

	public:
		constexpr WrapInline() = default;

		template<typename... Args>
		constexpr explicit WrapInline(Args &&...args): m_obj(ox::forward<Args>(args)...) {
		}

		ox::CStringView typeName() const noexcept override {
			return ox::ModelTypeName_v<T>;
		}

		int typeVersion() const noexcept override {
			return ox::ModelTypeVersion_v<T>;
		}

		constexpr T &obj() noexcept override {
			return m_obj;
		}

};

template<typename T, typename... Args>
constexpr ox::UPtr<Wrap> makeWrap(Args &&...args) noexcept {
	return ox::make_unique<WrapInline<T>>(ox::forward<Args>(args)...);
}

template<typename T>
constexpr T &wrapCast(Wrap &ptr) noexcept {
	return static_cast<WrapT<T>&>(ptr).obj();
}

class BaseConverter {
	public:
		constexpr virtual ~BaseConverter() noexcept = default;

		[[nodiscard]]
		constexpr virtual ox::StringView srcTypeName() const noexcept = 0;

		[[nodiscard]]
		constexpr virtual int srcTypeVersion() const noexcept = 0;

		[[nodiscard]]
		constexpr virtual bool srcMatches(ox::StringViewCR pSrcTypeName, int pSrcTypeVersion) const noexcept = 0;

		[[nodiscard]]
		constexpr virtual bool dstMatches(ox::StringViewCR dstTypeName, int dstTypeVersion) const noexcept = 0;

		virtual ox::Result<ox::UPtr<Wrap>> convertPtrToPtr(keel::Context &ctx, Wrap &src) const noexcept = 0;

		virtual ox::Result<ox::UPtr<Wrap>> convertBuffToPtr(
				keel::Context &ctx, ox::BufferView const&srcBuff) const noexcept = 0;

		[[nodiscard]]
		constexpr bool matches(
				ox::StringViewCR srcTypeName, int const srcTypeVersion,
				ox::StringViewCR dstTypeName, int const dstTypeVersion) const noexcept {
			return srcMatches(srcTypeName, srcTypeVersion)
			    && dstMatches(dstTypeName, dstTypeVersion);
		}

};

template<typename SrcType, typename DstType>
class Converter: public BaseConverter {
	public:
		[[nodiscard]]
		constexpr ox::StringView srcTypeName() const noexcept final {
			return ox::ModelTypeName_v<SrcType>;
		}

		[[nodiscard]]
		constexpr int srcTypeVersion() const noexcept final {
			return ox::ModelTypeVersion_v<SrcType>;
		}

		[[nodiscard]]
		constexpr bool srcMatches(ox::StringViewCR pSrcTypeName, int pSrcTypeVersion) const noexcept final {
			constexpr auto SrcTypeName = ox::requireModelTypeName<SrcType>();
			constexpr auto SrcTypeVersion = ox::requireModelTypeVersion<SrcType>();
			return pSrcTypeName == SrcTypeName
			    && pSrcTypeVersion == SrcTypeVersion;
		}

		[[nodiscard]]
		constexpr bool dstMatches(ox::StringViewCR dstTypeName, int dstTypeVersion) const noexcept final {
			constexpr auto DstTypeName = ox::StringView{ox::requireModelTypeName<DstType>()};
			constexpr auto DstTypeVersion = ox::requireModelTypeVersion<DstType>();
			return dstTypeName == DstTypeName
			    && dstTypeVersion == DstTypeVersion;
		}

		ox::Result<ox::UPtr<Wrap>> convertPtrToPtr(
				keel::Context &ctx, Wrap &src) const noexcept final {
			ox::Result<ox::UPtr<Wrap>> dst{makeWrap<DstType>()};
			OX_RETURN_ERROR(convert(ctx, wrapCast<SrcType>(src), wrapCast<DstType>(*dst.value)));
			return dst;
		}

		ox::Result<ox::UPtr<Wrap>> convertBuffToPtr(
				keel::Context &ctx, ox::BufferView const&srcBuff) const noexcept final {
			OX_REQUIRE_M(src, readAsset<SrcType>(srcBuff));
			ox::Result<ox::UPtr<Wrap>> dst{makeWrap<DstType>()};
			OX_RETURN_ERROR(convert(ctx, src, wrapCast<DstType>(*dst.value)));
			return dst;
		}

	protected:
		virtual ox::Error convert(keel::Context &ctx, SrcType&, DstType&) const noexcept = 0;

};

ox::Result<ox::UPtr<Wrap>> convert(
		keel::Context &ctx,
		ox::BufferView const&srcBuffer,
		ox::StringViewCR dstTypeName,
		int dstTypeVersion) noexcept;

ox::Result<ox::UPtr<Wrap>> convert(
		keel::Context &ctx,
		Wrap &src,
		ox::StringViewCR dstTypeName,
		int dstTypeVersion) noexcept;

ox::Result<ox::UPtr<Wrap>> convert(
		keel::Context &ctx,
		auto &src,
		ox::StringViewCR dstTypeName,
		int const dstTypeVersion) noexcept {
	WrapRef ref{src};
	return convert(ctx, static_cast<Wrap&>(ref), dstTypeName, dstTypeVersion);
}

ox::Result<ox::UPtr<Wrap>> convert(
		keel::Context &ctx,
		auto const&src,
		ox::StringViewCR dstTypeName,
		int const dstTypeVersion) noexcept {
	auto srcCpy = src;
	WrapRef ref{srcCpy};
	return convert(ctx, ref, dstTypeName, dstTypeVersion);
}

template<typename DstType>
ox::Result<DstType> convertObjToObj(
		keel::Context &ctx,
		auto &src) noexcept {
	OX_REQUIRE_M(out, convert(ctx, WrapRef{src}, ox::ModelTypeName_v<DstType>, ox::ModelTypeVersion_v<DstType>));
	return std::move(wrapCast(*out));
}

template<typename DstType>
ox::Result<DstType> convert(keel::Context &ctx, ox::BufferView const&src) noexcept {
	OX_REQUIRE(out, convert(ctx, src, ox::ModelTypeName_v<DstType>, ox::ModelTypeVersion_v<DstType>));
	return std::move(wrapCast<DstType>(out));
}

template<typename DstType>
ox::Error convert(keel::Context &ctx, ox::BufferView const&buff, DstType &outObj) noexcept {
	OX_REQUIRE(out, convert(ctx, buff, ox::ModelTypeName_v<DstType>, ox::ModelTypeVersion_v<DstType>));
	outObj = std::move(wrapCast<DstType>(*out));
	return {};
}

template<typename DstType>
ox::Error convertObjToObj(keel::Context &ctx, auto &src, DstType &outObj) noexcept {
	OX_REQUIRE(outPtr, convert(ctx, src, ox::ModelTypeName_v<DstType>, ox::ModelTypeVersion_v<DstType>));
	outObj = std::move(wrapCast<DstType>(*outPtr));
	return {};
}

template<typename DstType>
ox::Result<ox::Buffer> convertBuffToBuff(
		keel::Context &ctx, ox::BufferView const&src, ox::ClawFormat const fmt) noexcept {
	OX_REQUIRE(out, convert(ctx, src, ox::ModelTypeName_v<DstType>, ox::ModelTypeVersion_v<DstType>));
	return ox::writeClaw<DstType>(wrapCast<DstType>(*out), fmt);
}

template<typename From, typename To, ox::ClawFormat fmt = ox::ClawFormat::Metal>
ox::Result<bool> transformRule(keel::Context &ctx, ox::Buffer &buff, ox::StringViewCR typeId) noexcept {
	if (typeId == ox::ModelTypeId_v<From>) {
		OX_RETURN_ERROR(keel::convertBuffToBuff<To>(ctx, buff, fmt).moveTo(buff));
		return true;
	}
	return false;
};

}
